#ifdef _WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef __APPLE__
#include <GL/gl.h>
#include <GL/glut.h>
#else
#include <OpenGL/gl.h>
#include <GLUT/glut.h>
#endif
#include <AR/gsub.h>
#include <AR/video.h>
#include <AR/param.h>
#include <AR/ar.h>
#include "GLMetaseq.h"

#define MAX_PAIR_CONT 8

static char *conf_name = "data/mqoseq.conf";

typedef struct {
    char name[512];             /* filename */
    double center[2];           /* center */
    double width;               /* width of pattern */

    int id;                     /* pattern ID */
    int visible;                /* is visible? */
    double trans[3][4];         /* matrix */
} pattern_data;

typedef struct {
    int cnt;                    /* number of models */
    double scale;               /* scale */
    char basename[512];         /* base filename */

    MQO_SEQUENCE seq;           /* sequence data */
    int cur;                    /* current frame */
} model_data;

typedef struct {
    char vconf[512];            /* video config */

    char cparam_name[512];      /* filename of camera parameter */
    ARParam cparam;             /* camera parameter data */

    int thresh;                 /* threshold to detect */
    int pair_cnt;               /* number of marker/model pair */
    pattern_data patt[MAX_PAIR_CONT]; /* marker pattern data */
    model_data model[MAX_PAIR_CONT];  /* 3D model data */

    int width;                  /* width of video */
    int height;                 /* height of video */
} conf_data;

static conf_data conf;

static char *get_line(char *buf, int n, FILE *fp)
{
    char *ptr;
    size_t m;

    while(1) {
        ptr = fgets(buf, n, fp);

        if(ptr == NULL) {
            return ptr;
        }

        m = strlen(buf);
        while(m > 0 && (buf[m - 1] == '\n' || buf[m - 1] == '\r')) {
            buf[m - 1] = 0;
            m -= 1;
        }

        if(m > 0 && buf[0] != '#') {
            return ptr;
        }
    }
}

static int read_pattern_conf(FILE *fp, pattern_data *patt)
{
    char buf[512];

    /* read name of pattern file */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        return 0;
    }
    strcpy(patt->name, buf);

    /* read marker width */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        return 0;
    }
    if(sscanf(buf, "%lf", &patt->width) != 1) {
        return 0;
    }

    return 1;
}

static int read_model_conf(FILE *fp, model_data *model)
{
    char buf[512];

    /* read number of model files */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        return 0;
    }
    if(sscanf(buf, "%d", &model->cnt) != 1) {
        return 0;
    }

    /* read scale of model files */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        return 0;
    }
    if(sscanf(buf, "%lf", &model->scale) != 1) {
        return 0;
    }

    /* read base name of model files */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        return 0;
    }
    strcpy(model->basename, buf);

    return 1;
}

static int read_conf(char *fname)
{
    FILE *fp;
    char buf[512];
    int ret = 1;
    int i;

    memset(&conf, 0, sizeof(conf));

    fp = fopen(fname, "r");
    if(fp == NULL) {
        ret = 0;
        return ret;
    }

    /* read name of video conf */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        ret = 0;
        goto done;
    }
    strcpy(conf.vconf, buf);

    /* read name of camera param */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        ret = 0;
        goto done;
    }
    strcpy(conf.cparam_name, buf);

    /* read marker detect threshold */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        ret = 0;
        goto done;
    }
    if(sscanf(buf, "%d", &conf.thresh) != 1) {
        ret = 0;
        goto done;
    }

    /* read number of marker/model pair */
    if(get_line(buf, sizeof(buf), fp) == NULL) {
        ret = 0;
        goto done;
    }
    if(sscanf(buf, "%d", &conf.pair_cnt) != 1) {
        ret = 0;
        goto done;
    }
    if(conf.pair_cnt > MAX_PAIR_CONT) {
        conf.pair_cnt = MAX_PAIR_CONT;
    }

    /* read marker pattern and model */
    for(i = 0; i < conf.pair_cnt; i++) {
        if(read_pattern_conf(fp, &conf.patt[i]) == 0 ||
           read_model_conf(fp, &conf.model[i]) == 0) {
            ret = 0;
            goto done;
        }
    }

    ret = 1;

  done:
    if(fp != NULL) {
        fclose(fp);
    }

    return ret;
}

static int read_models(void)
{
    int i;

    for(i = 0; i < conf.pair_cnt; i++) {
        conf.model[i].seq = mqoCreateSequence(conf.model[i].basename,
                                              conf.model[i].cnt,
                                              conf.model[i].scale);
        conf.model[i].cur = 0;
    }

    return 1;
}

static void init(void)
{
    ARParam wparam;
    int i;

    /* read conf file */
    printf("Reading config file..."); fflush(stdout);
    if(read_conf(conf_name) == 0) {
        fprintf(stderr, "failed to read conf file: %s\n", conf_name);
        exit(1);
    }
    printf("done\n");

    /* open video */
    printf("Opening video..."); fflush(stdout);
    if(arVideoOpen(conf.vconf) < 0) {
        fprintf(stderr, "failed to open video: %s\n", conf.vconf);
        exit(1);
    }
    printf("done\n");

    if(arVideoInqSize(&conf.width, &conf.height) < 0) {
        fprintf(stderr, "failed to get video size\n");
        exit(1);
    }

    /* load camera param */
    printf("Loading camera parameters..."); fflush(stdout);
    if(arParamLoad(conf.cparam_name, 1, &wparam) < 0) {
        fprintf(stderr, "failed to load camera param: %s\n", conf.cparam_name);
        exit(1);
    }
    printf("done\n");

    /* set/init camera param */
    arParamChangeSize(&wparam, conf.width, conf.height, &conf.cparam);
    arInitCparam(&conf.cparam);

    /* read pattern file */
    printf("Loading pattern file..."); fflush(stdout);
    for(i = 0; i < conf.pair_cnt; i++) {
        conf.patt[i].id = arLoadPatt(conf.patt[i].name);
        if(conf.patt[i].id < 0) {
            fprintf(stderr, "failed to load pattern file: %s\n",
                    conf.patt[i].name);
            exit(1);
        }
    }
    printf("done\n");

    /* initialize graphics */
    argInit(&conf.cparam, 1.0, 0, 0, 0, 0);

    /* read model files */
    printf("Loading model files..."); fflush(stdout);
    if(read_models() == 0) {
        fprintf(stderr,
                "failed to load model files\n");
        exit(1);
    }
    printf("done\n");
}

static void cleanup(void)
{
    int i;

    for(i = 0; i < conf.pair_cnt; i++) {
        mqoDeleteSequence(conf.model[i].seq);
    }

    arVideoCapStop();
    arVideoClose();
    argCleanup();
}

static void key_event(unsigned char key, int x, int y)
{
    if(key == 0x1b) {
        /* ESC key */
        cleanup();
        exit(0);
    }
}

static void draw(void)
{
    double gl_trans[16];
    int i;

    GLfloat light_position[] = {100.0, -200.0, 200.0, 0.0};
    GLfloat light_ambient[] = {0.1, 0.1, 0.1, 0.1};
    GLfloat light_diffuse[] = {0.9, 0.9, 0.9, 0.1};

    argDrawMode3D();
    argDraw3dCamera(0, 0);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    for(i = 0; i < conf.pair_cnt; i++) {
        if(conf.patt[i].visible == 0) {
            continue;
        }

        argConvGlpara(conf.patt[i].trans, gl_trans);
        glMatrixMode(GL_MODELVIEW);
        glLoadMatrixd(gl_trans);

        glLightfv(GL_LIGHT0, GL_POSITION, light_position);
        glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);

        glMatrixMode(GL_MODELVIEW);

        glRotatef(90.0, 1.0, 0.0, 0.0);
        mqoCallSequence(conf.model[i].seq, conf.model[i].cur);

        conf.model[i].cur += 1;
        if(conf.model[i].cur >= conf.model[i].cnt) {
            conf.model[i].cur = 0;
        }
    }

    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
}

static void main_loop(void)
{
    ARUint8 *cdata;
    ARMarkerInfo *marker_info;
    int marker_num;
    int i, j, idx[MAX_PAIR_CONT];

    /* get video image */
    cdata = arVideoGetImage();
    if(cdata == NULL) {
        arUtilSleep(2);
        return;
    }

    /* clear draw buffer */
    glDrawBuffer(GL_BACK);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClearDepth(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* draw video image as background */
    argDrawMode2D();
    argDispImage(cdata, 0, 0);

    /* prepare to next frame */
    arVideoCapNext();

    /* detect marker from video image */
    if(arDetectMarker(cdata, conf.thresh, &marker_info, &marker_num) < 0) {
        fprintf(stderr, "failed to detect marker\n");
        cleanup();
        exit(0);
    }

    /* search marker */
    for(i = 0; i < conf.pair_cnt; i++) {
        idx[i] = -1;
    }

    for(i = 0; i < marker_num; i++) {
        for(j = 0; j < conf.pair_cnt; j++) {
            if(marker_info[i].id != conf.patt[j].id) {
                continue;
            }

            if(idx[j] < 0 ||
               marker_info[i].cf > marker_info[idx[j]].cf) {
                idx[j] = i;
            }
        }
    }

    for(i = 0; i < conf.pair_cnt; i++) {
        if(idx[i] >= 0) {
            /* get camera pos */
            conf.patt[i].visible = 1;
            arGetTransMat(&marker_info[idx[i]],
                          conf.patt[i].center, conf.patt[i].width,
                          conf.patt[i].trans);
        }
        else {
            conf.patt[i].visible = 0;
        }
    }

    /* draw objects */
    draw();

    /* swap buffers to show images */
    argSwapBuffers();
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    init();

    arVideoCapStart();
    argMainLoop(NULL, key_event, main_loop);

    return 0;
}
